/* numbers.h -- Copyright 1989, 1993 Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* ReadNumber and WriteNumber take/return a long, using a compression
 * algorithm to reduce the amount of data taken.
 *
 * They use (char *) pointers instead if prefixed with an s.
 *
 * $Id: numbers.h,v 1.6 93/12/08 23:09:48 lee Exp $
 *
 */

#ifdef HAVE_INLINE

#define PutC(ch, Sp, base, maxlen) \
    ((((*(S)) - (base)) >= maxlen) ? -1 : ( (*((*(Sp))++) = (char) (ch)), 0))

INLINE int
sWriteNumber(s, Number, base, maxlen)
    char **s;
    unsigned long Number;
    char *base;
    unsigned int maxlen;
{
    while (Number > 0177) {
	/* the 0200 means more data follows... */
	if (PutC((Number & 0177) | 0200, s, base, maxlen) < 0) {
	    return -1;
	}
	Number >>= 7;
    }
    if (PutC(Number & 0177, s, base, maxlen) < 0) {
	return -1;
    }
}


#define GetC(S, base, maxlen) \
    ( (unsigned int) * (unsigned char *) ((* (unsigned char **)S)++) )

INLINE unsigned long
sReadNumber(s)
    char **s;
{
    unsigned long Result = 0L;
    int ThereIsMore;
    int Shift = 0;

    do {
	Result |= ((ThereIsMore = GetC(s)) & 0177) << Shift;
	ThereIsMore &= 0200;
	Shift += 7;
    } while (ThereIsMore);
    return Result;
}

#else /* HAVE_INLINE */

extern unsigned long fReadNumber();
extern unsigned long sReadNumber();

extern void fWriteNumber();
extern int sWriteNumber();

#endif /* HAVE_INLINE */

