/* wpblock.c -- Copyright 1989, 1983 Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* $Id: wpblock.c,v 1.9 93/11/28 14:59:27 lee Exp $
 *
 * This file defines
 * unsigned long Putpblock(t_WordInfo *WordInfo; t_pblock *pblock);
 * 
 * unsigned long PutWordPlaces(
 *    t_WordPlace *WordPlaces;
 *    t_WID WID;
 *    unsigned char *Block;
 *    unsigned BlockLength;
 *    unsigned long NextOffset;
 *    unsigned long NextLength,
 *    unsigned long NumberToWrite;
 * );
 *
 * unsigned long AddWordPlaces(
 *    t_WordPlace *WordPlaces;
 *    t_WID WID;
 *    unsigned long Offset;
 *    unsigned long NumberToWrite;
 * );
 *
 */

#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#include <stdio.h> /* stderr, also for fileinfo.h */
#include <sys/types.h>
#include "fileinfo.h" /* for wordinfo.h */
#include "wordinfo.h"
#include "wordrules.h"
#include "pblock.h"
#include "putbyte.h"
#include "emalloc.h"

/** Unix system calls that need to be declared: **/

/** C library functions that need to be declared: **/

/** lqtext library functions that need to be declared: **/
extern void MkWIBH();
extern void SortWordPlaces();
extern void FlushBlock();
extern int _PutByte(), _PutLong();
extern unsigned long FindFreeBlock();

#ifdef ASCIITRACE
extern void fprintWordInfo(/*stream, W, Caller*/);
#endif

/** Functions within this file that need to be declared: **/
unsigned long PutWordPlaces();
/** **/


#ifdef ASCIITRACE
extern int AsciiTrace;
#endif

/* Write out an entire (presumably new) data entry, and
 * return a disk pointer to the start of the chain
 */
API unsigned long
Putpblock(WordInfo, pblock)
    t_WordInfo *WordInfo;
    t_pblock *pblock;
{
    unsigned long BlockLength;

    if (WordInfo->DataBlock) {
	(void) efree(WordInfo->DataBlock);
	WordInfo->DataBlock = (char *) 0;
    }

    WordInfo->Offset = pblock->ChainStart =
				FindFreeBlock(WordInfo->WID, &BlockLength);
    (void) MkWIBH(WordInfo, pblock);

    (void) PutWordPlaces(
	pblock->WordPlaces,
	WordInfo->WID,
	(unsigned long) 0L, /* 0 means 1st block not in data file */
	WordInfo->DataBlock,
	(unsigned char *) WordInfo->WordPlaceStart,
	(unsigned) WIDBLOCKSIZE,
	WordInfo->Offset,
	BlockLength,
	pblock->NumberOfWordPlaces
    );

#ifdef ASCIITRACE
    if (AsciiTrace & 0x10) {
	t_pblock *p2;
	register long place;
	extern t_pblock *Getpblock();

	p2 = Getpblock(WordInfo);

	if (p2->NumberOfWordPlaces != pblock->NumberOfWordPlaces) {
	    Error(E_FATAL|E_BUG,
		"%s: %d: places differ %ld vs. %ld",
		__FILE__, __LINE__,
		p2->NumberOfWordPlaces,
		pblock->NumberOfWordPlaces
	    );
	}

	for (place = 0; place < p2->NumberOfWordPlaces; place++) {
#define PCMP(p1,p2,place) \
((p2->WordPlaces[place].FID == p2->WordPlaces[place].FID) && \
(p2->WordPlaces[place].BlockInFile == p2->WordPlaces[place].BlockInFile) && \
(p2->WordPlaces[place].WordInBlock == p2->WordPlaces[place].WordInBlock) && \
(p2->WordPlaces[place].Flags == p2->WordPlaces[place].Flags) && \
(p2->WordPlaces[place].StuffBefore == p2->WordPlaces[place].StuffBefore))

	    if (!PCMP(pblock, p2, place)) {
		Error(E_BUG|E_FATAL,
		    "%s: %d: differ @%ld: %ld,%ld,%d,0%o,%d/%ld,%ld,%d,0%o,%d",
		    __FILE__, __LINE__,
		    place,
		    pblock->WordPlaces[place].FID,
		    pblock->WordPlaces[place].BlockInFile,
		    pblock->WordPlaces[place].WordInBlock,
		    pblock->WordPlaces[place].Flags,
		    pblock->WordPlaces[place].StuffBefore,

		    p2->WordPlaces[place].FID,
		    p2->WordPlaces[place].BlockInFile,
		    p2->WordPlaces[place].WordInBlock,
		    p2->WordPlaces[place].Flags,
		    p2->WordPlaces[place].StuffBefore
		);
	    }
	}

	(void) efree((char *) p2);
    }
#endif
    return WordInfo->Offset;
}

/** CAUTION - this comment might be out of date CAUTION *********************
 ** WordPlaces are now stored as sequences, as follows:
 **  FID*2 -- 1, 2, 3 (usually, for FID) or 4 bytes			1-5
 **  (very, very occasionaly a variable-size number may be 5 bytes long.)
 **   . the bottom bit in the stored number determines whether there
 **     is more than one FID to follow
 ** Number of following places (only if prev. bit was 1) -- 1 byte	0-1
 ** For each following entry:-
 **   . for each of the following places:
 **     Block In File (long, 1-5 bytes, usually 1)			1-5
 **     Word In Block -- always 1 byte					1-1
 **		the bottom bit of this says if there are flags
 **     Flags -- always 1 byte, if present				0-1
 **	    (flags stored only if different from previous entry)
 **	    Stuff Before -- 1 byte					0-1
 **	    (if there are no flags, there's no Stuff Before byte, and
 **     we use the default value of 1)
 **
 ** Hence:	each sub-place takes from 2 to 9 bytes;
 **		each Place sequence takes from 3
 **		to (4 + 1) + 255 * (2..7) bytes.
 **		In most (I guess > 7/10) cases, flags will be 0, and
 **		StuffBefore will be the default of 1.
 **
 ** In practice, though, we store the difference since the last block-in-file,
 ** and the difference since the last FID, so that the numbers are usually
 ** on the small side.
 **
 **	I am hoping, of course, that the extra information stored is
 ** worth while!
 **	It might be possible to coalesce WordInBlock and BlockInFile using
 ** delta modulation -- i.e., storing the increment from the previous.  In
 ** this case, a flag bit could mean that those two values each occupy a
 ** nibble in a single byte.  Or, I could use a single byte, like this:
 **	[a b c d e f g h]
 **	a == 1 --> (efgh) is word in block inc., (bcd is block in file inc)
 ** but I need to do some real measurements to figure out how best to save
 ** space.  It really is worth while keeping the format as simple as I can,
 ** as this speeds retrieval.
 **
 **/

API unsigned long
PutWordPlaces(WordPlaces, WID, LastStart, Block, DataStart, BlockLength, NextOffset, NextSize, NumberToWrite)
    t_WordPlace *WordPlaces;
    t_WID WID;
    unsigned long LastStart;
    unsigned char *Block;
    unsigned char *DataStart;
    unsigned BlockLength;
    unsigned long NextOffset;
    unsigned long NextSize;
    unsigned long NumberToWrite;
{
    unsigned char *q = DataStart;
    unsigned long L;
    int CurrentPlace = 0;
    t_FID LastFID = 0;
    unsigned long LastBlock = 0L;
    unsigned char LastFlags = 0;

#ifdef ALWAYSSORT
    /* Sort the pblock to simplify subsequent accesses,
     * and also to allow more space-efficient encoding, recording the change
     * (increment) since the previous FID or Block in the list, instead of
     * the actual number.  The WriteNumber package works much better if
     * most numbers are small.
     */
    if (NumberToWrite > 1) {
	SortWordPlaces(NumberToWrite, WordPlaces);
    }
#endif

    while (CurrentPlace < 0 || CurrentPlace < NumberToWrite) {
	unsigned long NumberOfRepeats;
	unsigned char U;
	t_FID FID = WordPlaces[CurrentPlace].FID;
	int LastPlace;

	if (FID == 0) {
	    Error(E_BUG, "PutWordPlaces WID %ld, FID %ld is Zero!",
		WID, CurrentPlace
	    );
	}

	/* Determine the number of Places in the same file;
	 * We use a long here.  Previous versions used a single byte,
	 * and if there were more than 255 occurrences of a word in the
	 * same file, used multiple sequences.  Multiple sequences can now
	 * also occur through updates to the index, and the two cases need
	 * to be distinguished, because in the update case the running
	 * totals (LastFID) are reset to 0.
	 */
	NumberOfRepeats = 0;
	LastPlace = CurrentPlace;
	while (LastPlace < NumberToWrite) {
	    if (WordPlaces[LastPlace].FID != FID) {
		break;
	    }
	    ++NumberOfRepeats;
	    ++LastPlace;
	}

	L = (FID - LastFID) << 1;
	if (NumberOfRepeats > 1) L |= 01L;

	if (L == 0L) {
	    Error(E_WARN|E_INTERNAL,
		"%s: %d: assertion failed: L is 0, can't tell if update",
		__FILE__, __LINE__
	    );
	    LastFID = 0L;
	    L = (FID << 1);
	    if (NumberOfRepeats > 1) L |= 01L;

	    /* fake the beginning of an update */
	    if (PutByte(0, WID, &q, &Block, &BlockLength,
				    &LastStart, &NextOffset, &NextSize)  < 0) {
		return CurrentPlace;
	    }
	}

	LastFID = FID;
	if (PutLong(L, WID, &q, &Block, &BlockLength,
				    &LastStart, &NextOffset, &NextSize) < 0) {
	    return CurrentPlace;
	}
	if (L & 01L) {
	    if (PutLong(NumberOfRepeats, WID, &q, &Block, &BlockLength,
				    &LastStart, &NextOffset, &NextSize)  < 0) {
		return CurrentPlace;
	    }
	}

	LastBlock = 0;

	for (; NumberOfRepeats != 0; --NumberOfRepeats) {
	    if (CurrentPlace > NumberToWrite) {
		Error(E_BUG,
		"Word %ld: Entry for file %lu has more matches than expected",
								WID, FID);
	    }
#ifdef ASCIITRACE
	    if (WordPlaces[CurrentPlace].BlockInFile < LastBlock) {
		Error(E_BUG,
		    "PutWordPlaces Sort WID %ld failed, backwards blocks",
		    WID
		);
	    } else if (CurrentPlace &&
		    (WordPlaces[CurrentPlace].FID ==
			    WordPlaces[CurrentPlace - 1].FID) &&
		    (WordPlaces[CurrentPlace].BlockInFile == LastBlock) &&
		    (WordPlaces[CurrentPlace].WordInBlock <=
			    WordPlaces[CurrentPlace - 1].WordInBlock)) {
		Error(E_BUG,
"PutWordPlaces Sort WID %ld failed, FID %ld: Blk %d: WIB %d <= %d",
		    WID, FID, LastBlock, WordPlaces[CurrentPlace].WordInBlock,
		    WordPlaces[CurrentPlace - 1].WordInBlock
		);
	    }
#endif /* ASCIITRACE */
#ifdef DEBUGPLACES
	    U = '{';
	    if (PutByte(U, WID, &q, &Block, &BlockLength,
				&LastStart, &NextOffset, &NextSize) < 0) {
		return CurrentPlace;
	    }
#endif
	    /* write the block number */
	    L = WordPlaces[CurrentPlace].BlockInFile - LastBlock;
	    LastBlock += L;

	    if (PutLong(L, WID, &q, &Block, &BlockLength,
				&LastStart, &NextOffset, &NextSize) < 0) {
		return CurrentPlace;
	    }
	    U = (WordPlaces[CurrentPlace].WordInBlock << 1);
	    if (WordPlaces[CurrentPlace].StuffBefore != 1) {
		WordPlaces[CurrentPlace].Flags |= WPF_HASSTUFFBEFORE;
	    }
	    if (WordPlaces[CurrentPlace].Flags != LastFlags) {
		U |= 01;
	    }

	    if (U > 255) {
		Error(E_BUG, "WID %lu: WordInBlock (0%o) from FID %lu too big",
			WID, U, FID);
	    }

	    if (PutByte(U, WID, &q, &Block, &BlockLength,
				&LastStart, &NextOffset, &NextSize) < 0) {
		return CurrentPlace;
	    }
	    if (U & 01) {
		LastFlags = WordPlaces[CurrentPlace].Flags;
		if (PutByte(LastFlags, WID, &q, &Block, &BlockLength,
				&LastStart, &NextOffset, &NextSize) < 0) {
		    return CurrentPlace;
		}
	    }

	    /* Even if there are flags, there still might not be a separate
	     * entry for the number of preceding skipped bytes.
	     */
	    if (WordPlaces[CurrentPlace].Flags & WPF_HASSTUFFBEFORE) {
		if (PutByte(WordPlaces[CurrentPlace].StuffBefore, WID, &q,
			&Block, &BlockLength,
				&LastStart, &NextOffset, &NextSize) < 0) {
		    return CurrentPlace;
		}
	    }

#ifdef DEBUGPLACES
	    U = '}';
	    if (PutByte(U, WID, &q, &Block, &BlockLength,
				&LastStart, &NextOffset, &NextSize) < 0) {
		return CurrentPlace;
	    }
#endif
	    ++CurrentPlace;
	}
	if (CurrentPlace > LastPlace) {
	    Error(E_BUG, "PutWordPlaces: CurrentPlace %ld > LastPlace %ld",
						    CurrentPlace, LastPlace);
	}
    }

    if (LastStart) {
	/* NextOffset had better not be non-zero, but FlushBlock will
	 * take care of it (we have wasted a block in that case!).
	 * LastStart is zero if we fitted it all inside the WordInfo
	 * block.
	 */
	FlushBlock(
	    (char *) Block,
	    q - Block,
	    &NextOffset,
	    &LastStart,
	    WID
	);
    }
    return NumberToWrite;
}

API unsigned long
AddWordPlaces(WordPlaces, WID, Offset, NumberToWrite)
    t_WordPlace *WordPlaces;
    t_WID WID;
    unsigned long Offset;
    unsigned long NumberToWrite;
{
    extern unsigned char *LastBlockInChain();
    extern void UpdateMatchCount();

    unsigned char *Block;
    unsigned char *p;
    unsigned BlockLength;
    unsigned long WrittenPlaces;

    Block = LastBlockInChain(WID, &Offset, &p, &BlockLength);

    if (!Block) {
	/* failed for some reason, make a new entry */
	Error(E_BUG|E_WARN, "AddWordPlaces: LastBlockInChain failed");
	return 0L;
    }

    /* put a null byte to reset the File Counter when reading back the
     * word places.
     */
    if (*p != 0xFF) {
	Error(E_BUG|E_FATAL,
	    "%s: %d: AddWordPlaces WID %ld lastblock %ld, *p is %d (0%o), not 255 (0377)",
	    __FILE__, __LINE__, WID, Offset, *p, *p
	);
    }

    *p = '\0';
    ++p;

    /* We need at least 4 bytes to store a wordplace, so let's
     * make sure it's possible... the bytes are
     * TotalPlaces, FID, number in file (opt), BlockInFile, WordInBlock
     * 1		 1		       1	    1
     * In practice we generally need more...
     * However, because we compress the data, we often need only 2 bytes
     * for each place after that.
     */
    if (BlockLength >= 2 * NumberToWrite) {
	/* First try to fit them all in the same block: */
	WrittenPlaces = PutWordPlaces(
	    WordPlaces,
	    WID,
	    Offset, /* block to flush */
	    Block,
	    p,
	    BlockLength,
	    0L, /* no next offset */
	    0L, /* no next blocklength */
	    NumberToWrite
	);
    } else {
	WrittenPlaces = 0L;
    }

    if (WrittenPlaces != NumberToWrite) {
	unsigned NextLength;
	unsigned long NextBlock = FindFreeBlock(WID, &NextLength);

	/* We needed to use extra blocks, they didn't all fit in
	 * the one that was already partly in use.
	 */

	if (p[-1] != 0) {
	    Error(E_BUG|E_FATAL,
		"%s: %d: AddWordPlaces WID %ld lastblock %ld, p[-1] is %d (0%0), not 255 (0377)",
		__FILE__, __LINE__, WID, Offset, p[-1], p[-1]
	    );
	}

	WrittenPlaces = PutWordPlaces(
	    WordPlaces,
	    WID,
	    Offset,
	    Block,
	    p,
	    BlockLength,
	    NextBlock,
	    NextLength,
	    NumberToWrite
	);

	if (WrittenPlaces != NumberToWrite) {
	    Error(E_BUG,
		"%S; %d: didn't write places for %s, wrote %ld, not %ld",
		__FILE__, __LINE__, WrittenPlaces, NumberToWrite
	    );
	}
    }

    /* Now we need to update the index entry */
    UpdateMatchCount(WID, NumberToWrite);

    return NumberToWrite;
}
